@@ -19,7 +19,8 @@ __PACKAGE__->mk_accessors(
     qw/tree dispatch_types registered_dispatch_types
       method_action_class action_container_class
       preload_dispatch_types postload_dispatch_types
-    /
+      action_hash container_hash
+      /
 );
 
 # Preload these action types
@@ -38,20 +39,35 @@ See L<Catalyst>.
 
 =head1 DESCRIPTION
 
+This is the class that maps public urls to actions in your Catalyst
+application based on the attributes you set.
+
 =head1 METHODS
 
+=head2 new 
+
+Construct a new dispatcher.
+
 =cut
 
 sub new {
-    my $self = shift;
+    my $self  = shift;
     my $class = ref($self) || $self;
-                                    
-    my $obj = $class->SUPER::new( @_ );
-                                       
+
+    my $obj = $class->SUPER::new(@_);
+
     # set the default pre- and and postloads
-    $obj->preload_dispatch_types(  \@PRELOAD );        
+    $obj->preload_dispatch_types( \@PRELOAD );
     $obj->postload_dispatch_types( \@POSTLOAD );
-    return $obj;                        
+    $obj->action_hash(    {} );
+    $obj->container_hash( {} );
+
+    # Create the root node of the tree
+    my $container =
+      Catalyst::ActionContainer->new( { part => '/', actions => {} } );
+    $obj->tree( Tree::Simple->new( $container, Tree::Simple->ROOT ) );
+
+    return $obj;
 }
 
 =head2 $self->preload_dispatch_types
@@ -76,6 +92,8 @@ it with a C<+>, like so:
 
 =head2 $self->detach( $c, $command [, \@arguments ] )
 
+Documented in L<Catalyst>
+
 =cut
 
 sub detach {
@@ -86,6 +104,10 @@ sub detach {
 
 =head2 $self->dispatch($c)
 
+Delegate the dispatch to the action that matched the url, or return a
+message about unknown resource
+
+
 =cut
 
 sub dispatch {
@@ -106,106 +128,119 @@ sub dispatch {
 
 =head2 $self->forward( $c, $command [, \@arguments ] )
 
+Documented in L<Catalyst>
+
 =cut
 
 sub forward {
-    my $self    = shift;
-    my $c       = shift;
-    my $command = shift;
+    my ( $self, $c, $command ) = splice( @_, 0, 3 );
 
     unless ($command) {
         $c->log->debug('Nothing to forward to') if $c->debug;
         return 0;
     }
 
-    my $local_args = 0;
-    my $arguments  = $c->req->args;
-    if ( ref( $_[-1] ) eq 'ARRAY' ) {
-        $arguments  = pop(@_);
-        $local_args = 1;
+    my $args = [ @{ $c->request->arguments } ];
+
+    @$args = @{ pop @_ } if ( ref( $_[-1] ) eq 'ARRAY' );
+
+    my $action = $self->_invoke_as_path( $c, $command, $args )
+      || $self->_invoke_as_component( $c, $command, shift );
+
+    unless ($action) {
+        my $error =
+            qq/Couldn't forward to command "$command": /
+          . qq/Invalid action or component./;
+        $c->error($error);
+        $c->log->debug($error) if $c->debug;
+        return 0;
     }
 
-    my $result;
+    #push @$args, @_;
 
-    unless ( ref $command ) {
-        my $command_copy = $command;
-
-        unless ( $command_copy =~ s/^\/// ) {
-            my $namespace = $c->stack->[-1]->namespace;
-            $command_copy = "${namespace}/${command}";
-        }
-
-        unless ( $command_copy =~ /\// ) {
-            $result = $c->get_action( $command_copy, '/' );
-        }
-        else {
-            my @extra_args;
-          DESCEND: while ( $command_copy =~ s/^(.*)\/(\w+)$/$1/ ) {
-                my $tail = $2;
-                $result = $c->get_action( $tail, $1 );
-                if ($result) {
-                    $local_args = 1;
-                    $command    = $tail;
-                    unshift( @{$arguments}, @extra_args );
-                    last DESCEND;
-                }
-                unshift( @extra_args, $tail );
-            }
-        }
-    }
-
-    unless ($result) {
-
-        my $class = ref($command)
-          || ref( $c->component($command) )
-          || $c->component($command);
-        my $method = shift || 'process';
-
-        unless ($class) {
-            my $error =
-qq/Couldn't forward to command "$command". Invalid action or component./;
-            $c->error($error);
-            $c->log->debug($error) if $c->debug;
-            return 0;
-        }
-
-        if ( my $code = $class->can($method) ) {
-            my $action = $self->method_action_class->new(
-                {
-                    name      => $method,
-                    code      => $code,
-                    reverse   => "$class->$method",
-                    class     => $class,
-                    namespace => Catalyst::Utils::class2prefix(
-                        $class, $c->config->{case_sensitive}
-                    ),
-                }
-            );
-            $result = $action;
-        }
-
-        else {
-            my $error =
-              qq/Couldn't forward to "$class". Does not implement "$method"/;
-            $c->error($error);
-            $c->log->debug($error)
-              if $c->debug;
-            return 0;
-        }
-
-    }
-
-    if ($local_args) {
-        local $c->request->{arguments} = [ @{$arguments} ];
-        $result->execute($c);
-    }
-    else { $result->execute($c) }
+    local $c->request->{arguments} = $args;
+    $action->execute($c);
 
     return $c->state;
 }
 
+sub _action_rel2abs {
+    my ( $self, $c, $path ) = @_;
+
+    unless ( $path =~ m#^/# ) {
+        my $namespace = $c->stack->[-1]->namespace;
+        $path = "$namespace/$path";
+    }
+
+    $path =~ s#^/##;
+    return $path;
+}
+
+sub _invoke_as_path {
+    my ( $self, $c, $rel_path, $args ) = @_;
+
+    return if ref $rel_path;    # it must be a string
+
+    my $path = $self->_action_rel2abs( $c, $rel_path );
+
+    my ( $tail, @extra_args );
+    while ( ( $path, $tail ) = ( $path =~ m#^(?:(.*)/)?(\w+)?$# ) )
+    {                           # allow $path to be empty
+        if ( my $action = $c->get_action( $tail, $path ) ) {
+            push @$args, @extra_args;
+            return $action;
+        }
+        else {
+            return
+              unless $path
+              ; # if a match on the global namespace failed then the whole lookup failed
+        }
+
+        unshift @extra_args, $tail;
+    }
+}
+
+sub _find_component_class {
+    my ( $self, $c, $component ) = @_;
+
+    return ref($component)
+      || ref( $c->component($component) )
+      || $c->component($component);
+}
+
+sub _invoke_as_component {
+    my ( $self, $c, $component, $method ) = @_;
+
+    my $class = $self->_find_component_class( $c, $component ) || return 0;
+    $method ||= "process";
+
+    if ( my $code = $class->can($method) ) {
+        return $self->method_action_class->new(
+            {
+                name      => $method,
+                code      => $code,
+                reverse   => "$class->$method",
+                class     => $class,
+                namespace => Catalyst::Utils::class2prefix(
+                    $class, $c->config->{case_sensitive}
+                ),
+            }
+        );
+    }
+    else {
+        my $error =
+          qq/Couldn't forward to "$class". Does not implement "$method"/;
+        $c->error($error);
+        $c->log->debug($error)
+          if $c->debug;
+        return 0;
+    }
+}
+
 =head2 $self->prepare_action($c)
 
+Find an dispatch type that matches $c->req->path, and set args from it.
+
 =cut
 
 sub prepare_action {
@@ -244,21 +279,29 @@ sub prepare_action {
 
 =head2 $self->get_action( $action, $namespace )
 
+returns a named action from a given namespace.
+
 =cut
 
 sub get_action {
     my ( $self, $name, $namespace ) = @_;
     return unless $name;
-    $namespace ||= '';
-    $namespace = '' if $namespace eq '/';
 
-    my @match = $self->get_containers($namespace);
+    $namespace = join( "/", grep { length } split '/', $namespace || "" );
 
-    return unless @match;
+    return $self->action_hash->{"$namespace/$name"};
+}
 
-    if ( my $action = $match[-1]->get_action($name) ) {
-        return $action if $action->namespace eq $namespace;
-    }
+=head2 $self->get_action_by_path( $path );
+
+returns the named action by it's full path.
+
+=cut
+
+sub get_action_by_path {
+    my ( $self, $path ) = @_;
+    $path = "/$path" unless $path =~ /\//;
+    $self->action_hash->{$path};
 }
 
 =head2 $self->get_actions( $c, $action, $namespace )
@@ -268,8 +311,8 @@ sub get_action {
 sub get_actions {
     my ( $self, $c, $action, $namespace ) = @_;
     return [] unless $action;
-    $namespace ||= '';
-    $namespace = '' if $namespace eq '/';
+
+    $namespace = join( "/", grep { length } split '/', $namespace || "" );
 
     my @match = $self->get_containers($namespace);
 
@@ -278,48 +321,32 @@ sub get_actions {
 
 =head2 $self->get_containers( $namespace )
 
+Return all the action containers for a given namespace, inclusive
+
 =cut
 
 sub get_containers {
     my ( $self, $namespace ) = @_;
+    $namespace ||= '';
+    $namespace = '' if $namespace eq '/';
 
-    # If the namespace is / just return the root ActionContainer
+    my @containers;
 
-    return ( $self->tree->getNodeValue )
-      if ( !$namespace || ( $namespace eq '/' ) );
+    do {
+        push @containers, $self->container_hash->{$namespace};
+    } while ( $namespace =~ s#/[^/]+$## );
 
-    # Use a visitor to recurse down the tree finding the ActionContainers
-    # for each namespace in the chain.
+    return reverse grep { defined } @containers, $self->container_hash->{''};
 
-    my $visitor = Tree::Simple::Visitor::FindByPath->new;
-    my @path = split( '/', $namespace );
-    $visitor->setSearchPath(@path);
-    $self->tree->accept($visitor);
-
-    my @match = $visitor->getResults;
-    @match = ( $self->tree ) unless @match;
-
-    if ( !defined $visitor->getResult ) {
-
-        # If we don't manage to match, the visitor doesn't return the last
-        # node is matched, so foo/bar/baz would only find the 'foo' node,
-        # not the foo and foo/bar nodes as it should. This does another
-        # single-level search to see if that's the case, and the 'last unless'
-        # should catch any failures - or short-circuit this if this *is* a
-        # bug in the visitor and gets fixed.
-
-        if ( my $extra = $path[ ( scalar @match ) - 1 ] ) {
-            $visitor->setSearchPath($extra);
-            $match[-1]->accept($visitor);
-            push( @match, $visitor->getResult ) if defined $visitor->getResult;
-        }
-    }
-
-    return map { $_->getNodeValue } @match;
+    my @parts = split '/', $namespace;
 }
 
 =head2 $self->register( $c, $action )
 
+Make sure all required dispatch types for this action are loaded, then
+pass the action to our dispatch types so they can register it if required.
+Also, set up the tree with the action containers.
+
 =cut
 
 sub register {
@@ -347,38 +374,47 @@ sub register {
     return unless $reg + $priv;
 
     my $namespace = $action->namespace;
-    my $parent    = $self->tree;
-    my $visitor   = Tree::Simple::Visitor::FindByPath->new;
+    my $name      = $action->name;
 
-    if ($namespace) {
-        for my $part ( split '/', $namespace ) {
-            $visitor->setSearchPath($part);
-            $parent->accept($visitor);
-            my $child = $visitor->getResult;
-
-            unless ($child) {
-
-                # Create a new tree node and an ActionContainer to form
-                # its value.
-
-                my $container =
-                  Catalyst::ActionContainer->new(
-                    { part => $part, actions => {} } );
-                $child = $parent->addChild( Tree::Simple->new($container) );
-                $visitor->setSearchPath($part);
-                $parent->accept($visitor);
-                $child = $visitor->getResult;
-            }
-
-            $parent = $child;
-        }
-    }
+    my $container = $self->find_or_create_action_container($namespace);
 
     # Set the method value
-    $parent->getNodeValue->actions->{ $action->name } = $action;
+    $container->add_action($action);
+
+    $self->action_hash->{"$namespace/$name"} = $action;
+    $self->container_hash->{$namespace} = $container;
 }
 
-=head2 $self->setup_actions( $class, $component )
+sub find_or_create_action_container {
+    my ( $self, $namespace ) = @_;
+
+    my $tree ||= $self->tree;
+
+    return $tree->getNodeValue unless $namespace;
+
+    my @namespace = split '/', $namespace;
+    return $self->_find_or_create_namespace_node( $tree, @namespace )
+      ->getNodeValue;
+}
+
+sub _find_or_create_namespace_node {
+    my ( $self, $parent, $part, @namespace ) = @_;
+
+    return $parent unless $part;
+
+    my $child =
+      ( grep { $_->getNodeValue->part eq $part } $parent->getAllChildren )[0];
+
+    unless ($child) {
+        my $container = Catalyst::ActionContainer->new($part);
+        $parent->addChild( $child = Tree::Simple->new($container) );
+    }
+
+    $self->_find_or_create_namespace_node( $child, @namespace );
+}
+
+=head2 $self->setup_actions( $class, $context )
+
 
 =cut
 
@@ -390,33 +426,15 @@ sub setup_actions {
     $self->method_action_class('Catalyst::Action');
     $self->action_container_class('Catalyst::ActionContainer');
 
-    # Preload action types
-    for my $type ( @{$self->preload_dispatch_types} ) {
-        my $class = ($type =~ /^\+(.*)$/) ? $1 : "Catalyst::DispatchType::${type}";
-        eval "require $class";
-        Catalyst::Exception->throw( message => qq/Couldn't load "$class"/ )
-          if $@;
-        push @{ $self->dispatch_types }, $class->new;
-        $self->registered_dispatch_types->{$class} = 1;
-    }
-
-    # We use a tree
-    my $container =
-      Catalyst::ActionContainer->new( { part => '/', actions => {} } );
-    $self->tree( Tree::Simple->new( $container, Tree::Simple->ROOT ) );
+    my @classes =
+      $self->do_load_dispatch_types( @{ $self->preload_dispatch_types } );
+    @{ $self->registered_dispatch_types }{@classes} = (1) x @classes;
 
     foreach my $comp ( values %{ $c->components } ) {
         $comp->register_actions($c) if $comp->can('register_actions');
     }
 
-    # Postload action types
-    for my $type ( @{$self->postload_dispatch_types} ) {
-        my $class = ($type =~ /^\+(.*)$/) ? $1 : "Catalyst::DispatchType::${type}";
-        eval "require $class";
-        Catalyst::Exception->throw( message => qq/Couldn't load "$class"/ )
-          if $@;
-        push @{ $self->dispatch_types }, $class->new;
-    }
+    $self->do_load_dispatch_types( @{ $self->postload_dispatch_types } );
 
     return unless $c->debug;
 
@@ -453,6 +471,26 @@ sub setup_actions {
     $_->list($c) for @{ $self->dispatch_types };
 }
 
+sub do_load_dispatch_types {
+    my ( $self, @types ) = @_;
+
+    my @loaded;
+
+    # Preload action types
+    for my $type (@types) {
+        my $class =
+          ( $type =~ /^\+(.*)$/ ) ? $1 : "Catalyst::DispatchType::${type}";
+        eval "require $class";
+        Catalyst::Exception->throw( message => qq/Couldn't load "$class"/ )
+          if $@;
+        push @{ $self->dispatch_types }, $class->new;
+
+        push @loaded, $class;
+    }
+
+    return @loaded;
+}
+
 =head1 AUTHOR
 
 Sebastian Riedel, C<sri@cpan.org>
